import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

public class Orga
   extends JFrame
{
   private static final String
      VERSION = "0.001",
      TITRE_DEFAUT = "Orga v." + VERSION;
   private static final int
      LARGEUR_DEFAUT = 640,
      HAUTEUR_DEFAUT = 480;

   private static final int
      FORME_RECTANGLE = 0,
      FORME_LOSANGE = 1,
      FORME_ELLIPSE = 2;
   private int m_Forme;
   private static final int
      MODE_SELECTION = 0,
      MODE_CREATION = 1,
      MODE_DEPLACEMENT = 2,
      MODE_PLACERDEVANT = 3,
      MODE_PLACERDERRIERE = 4;
   private int m_Mode;

   private BarreEtat m_BarreEtat;
   private BarreOutilsFormes m_BarreOutilsFormes;
   private BarreOutilsActions m_BarreOutilsActions;

   private interface IAffichable
   {
      public abstract void Selectionner ();
      public abstract void Deselectionner ();
      public abstract boolean GetSelectionne();
      public abstract void Survoler ();
      public abstract void Desurvoler ();
      public abstract boolean GetSurvole();
      public abstract void Dessiner (Graphics g);
      public abstract boolean Dedans (Point p);
      public abstract void Translation (int dx, int dy);
      public abstract java.awt.geom.Rectangle2D GetBoundingBox ();
   }

   private class Polygone extends Polygon implements IAffichable
   {
      private boolean m_EstSelectionne;
      private boolean m_EstSurvole;
      public Polygone (int [] px, int [] py, final int nPoints)
      {
         super (px, py, nPoints);
         m_EstSelectionne = false;
         m_EstSurvole = false;
      }
      public void Selectionner()
      {
         m_EstSelectionne = true;
      }
      public void Deselectionner()
      {
         m_EstSelectionne = false;
      }
      public void Survoler()
      {
         m_EstSurvole = true;
      }
      public void Desurvoler()
      {
         m_EstSurvole = false;
      }
      public boolean GetSelectionne()
      {
         return m_EstSelectionne;
      }
      public boolean GetSurvole()
      {
         return m_EstSurvole;
      }
      public void Dessiner (Graphics g)
      {
         g.setColor (Color.black);
         g.drawPolygon (this);
         if (GetSurvole ())
            g.setColor (Color.blue);
         else if (GetSelectionne ())
            g.setColor (Color.yellow);
         else
            g.setColor (Color.green);
         g.fillPolygon (this);
      }
      public boolean Dedans (Point p)
      {
         return contains (p);
      }
      public void Translation (int dx, int dy)
      {
         translate (dx, dy);
      }
      public java.awt.geom.Rectangle2D GetBoundingBox ()
      {
         return getBounds ();
      }
   }

   private class Ovale extends java.awt.geom.Ellipse2D.Double implements IAffichable
   {
      private boolean m_EstSelectionne;
      private boolean m_EstSurvole;
      public Ovale (final int x1, final int y1, final int x2, final int y2)
      {
         super (x1, y1, x2 - x1, y2 - y1);
         m_EstSelectionne = false;
         m_EstSurvole = false;
      }
      public void Selectionner()
      {
         m_EstSelectionne = true;
      }
      public void Deselectionner()
      {
         m_EstSelectionne = false;
      }
      public void Survoler()
      {
         m_EstSurvole = true;
      }
      public void Desurvoler()
      {
         m_EstSurvole = false;
      }
      public boolean GetSelectionne()
      {
         return m_EstSelectionne;
      }
      public boolean GetSurvole()
      {
         return m_EstSurvole;
      }
      public void Dessiner (Graphics g)
      {
         g.setColor (Color.black);
         g.drawOval ((int) getX (), (int) getY (), (int) getWidth (), (int) getHeight ());
         if (GetSurvole ())
            g.setColor (Color.blue);
         else if (GetSelectionne ())
            g.setColor (Color.yellow);
         else
            g.setColor (Color.green);
         g.fillOval ((int) getX (), (int) getY (), (int) getWidth (), (int) getHeight ());
      }
      public boolean Dedans (Point p)
      {
         return contains (p);
      }
      public void Translation (int dx, int dy)
      {
         setFrame (getX () + dx, getY () + dy, getHeight (), getWidth ());
      }
      public java.awt.geom.Rectangle2D GetBoundingBox ()
      {
         return getBounds2D ();
      }
   }

   private IAffichable [] m_Objets;
   private IAffichable [] m_Selection;
   private IAffichable m_Survol;
   private Canevas m_Canevas;

   private void AjouterObjet(IAffichable p)
   {
      IAffichable [] Temp = new IAffichable [m_Objets.length + 1];
      for (int i = 0; i < m_Objets.length; i++) Temp[i] = m_Objets[i];
      Temp[m_Objets.length] = p;
      m_Objets = Temp;
   }

   private void CreerForme (final Point Debut, final Point Fin)
   {
      switch (m_Forme)
      {
      case FORME_LOSANGE:
      {
         int [] px = new int [4];
         int [] py = new int [4];
         final int XMoyen = (int) ((Debut.getX () + Fin.getX ()) / 2);
         final int YMoyen = (int) ((Debut.getY () + Fin.getY ()) / 2);
         px[0] = XMoyen ;             py[0] = (int) Debut.getY ();
         px[1] = (int) Debut.getX (); py[1] = YMoyen;
         px[2] = XMoyen;              py[2] = (int) Fin.getY ();
         px[3] = (int) Fin.getX ();   py[3] = YMoyen;
         AjouterObjet (new Polygone (px, py, 4));
      }  break;
      case FORME_RECTANGLE:
      {
         int [] px = new int [4];
         int [] py = new int [4];
         px[0] = (int) Debut.getX (); py[0] = (int) Debut.getY ();
         px[1] = (int) Debut.getX (); py[1] = (int) Fin.getY ();
         px[2] = (int) Fin.getX ();   py[2] = (int) Fin.getY ();
         px[3] = (int) Fin.getX ();   py[3] = (int) Debut.getY ();
         AjouterObjet (new Polygone (px, py, 4));
      }  break;
      case FORME_ELLIPSE:
      {
         final double MinX = Math.min (Debut.getX (), Fin.getX ());
         final double MinY = Math.min (Debut.getY (), Fin.getY ());
         final double MaxX = Math.max (Debut.getX (), Fin.getX ());
         final double MaxY = Math.max (Debut.getY (), Fin.getY ());
         AjouterObjet
            (new Ovale ((int) MinX, (int) MinY, (int) (MaxX), (int) (MaxY)));
      }  break;
      default:
      }
   }

   private void MiseAJourZone (final Point Debut, final Point Fin)
   {
      int [] px = new int [4];
      int [] py = new int [4];
      px[0] = (int) Debut.getX (); py[0] = (int) Debut.getY ();
      px[1] = (int) Debut.getX (); py[1] = (int) Fin.getY ();
      px[2] = (int) Fin.getX ();   py[2] = (int) Fin.getY ();
      px[3] = (int) Fin.getX ();   py[3] = (int) Debut.getY ();
      m_Canevas.SetZone (new Polygon (px, py, 4));
   }

   private void MiseAJourPosition (Point posPointeur)
   {
      IAffichable p = null;
      for (int i = m_Objets.length; --i >= 0 && p == null;)
      {
         if (m_Objets[i] != null &&
             m_Objets[i].Dedans (posPointeur))
            p = m_Objets[i];
      }
      if (m_Selection == null && p != m_Survol)
      {
         System.err.print ("ICI:");
         if (m_Selection == null) System.err.print (" (m_Selection == null)");
         if (p != m_Survol) System.err.print (" (p != m_Survol)");
         System.err.println ("");
         if (p != null)
            p.Survoler();
         if (m_Survol != null)
            m_Survol.Desurvoler();
         m_Survol = p;
         m_Canevas.repaint ();
      }
   }

   //
   // Gestionnaire de souris
   //
   private class GestionnaireSouris
      implements MouseListener, MouseMotionListener
   {
      private Point m_Depart, m_Fin;
      private Point m_DerniereTranslation;

      public GestionnaireSouris ()
      {
         m_DerniereTranslation = new Point (0, 0);
      }

      public void mouseDragged (MouseEvent me)
      {
         m_Fin = me.getPoint ();
         if (m_Mode == MODE_DEPLACEMENT)
         {
            if (m_Survol != null)
            {
               m_Survol.Translation((int) -m_DerniereTranslation.getX (),
                                    (int) -m_DerniereTranslation.getY ());
               m_DerniereTranslation =
                  new Point ((int) (m_Fin.getX () - m_Depart.getX ()),
                             (int )(m_Fin.getY () - m_Depart.getY ()));
               m_Survol.Translation((int) m_DerniereTranslation.getX (),
                                    (int) m_DerniereTranslation.getY ());
            }
            else if (m_Selection != null)
            {
               for (int i = 0; i < m_Selection.length; i++)
               {
                  if (m_Selection[i] != null)
                  {
                     m_Selection[i].Translation((int) -m_DerniereTranslation.getX (),
                                                (int) -m_DerniereTranslation.getY ());
                     m_DerniereTranslation =
                        new Point ((int) (m_Fin.getX () - m_Depart.getX ()),
                                   (int )(m_Fin.getY () - m_Depart.getY ()));
                     m_Selection[i].Translation((int) m_DerniereTranslation.getX (),
                                                (int) m_DerniereTranslation.getY ());
                  }
               }
            }
         }
         else
            MiseAJourZone (m_Depart, m_Fin);
         m_Canevas.repaint ();
      }

      public void mouseMoved (MouseEvent me)
      {
         MiseAJourPosition (me.getPoint ());
      }

      public void mousePressed (MouseEvent me)
      {
         m_Depart = me.getPoint ();
         m_DerniereTranslation = new Point (0, 0);
/*
         if (me.isMetaDown ())
            m_Mode = MODE_LOSANGE;
         else
            m_Mode = MODE_RECTANGLE;
            */
      }
      public void mouseReleased (MouseEvent me)
      {
         m_Fin = me.getPoint ();
         switch (m_Mode)
         {
         case MODE_SELECTION:
            break;
         case MODE_CREATION:
            CreerForme (m_Depart, m_Fin);
            break;
         case MODE_DEPLACEMENT:
            break;
         case MODE_PLACERDEVANT:
            break;
         case MODE_PLACERDERRIERE:
            break;
         default:
         };
         m_Canevas.SetZone (null);
         m_Canevas.repaint ();
      }
      public void mouseClicked (MouseEvent me)
      {
         switch (m_Mode)
         {
         case MODE_SELECTION:
            break;
         case MODE_CREATION:
            break;
         case MODE_DEPLACEMENT:
            break;
         case MODE_PLACERDEVANT:
            if (m_Survol != null)
            {
               int i;
               for (i = 0; i < m_Objets.length && m_Objets[i] != m_Survol; i++);
               IAffichable Temp = m_Objets[m_Objets.length - 1];
               m_Objets[m_Objets.length - 1] = m_Objets[i];
               m_Objets[i] = Temp;
            }
            m_Canevas.SetZone (null);
            MiseAJourPosition (me.getPoint ());
            m_Canevas.repaint ();
            break;
         case MODE_PLACERDERRIERE:
            if (m_Survol != null)
            {
               int i;
               for (i = 0; i < m_Objets.length && m_Objets[i] != m_Survol; i++);
               IAffichable Temp = m_Objets[0];
               m_Objets[0] = m_Objets[i];
               m_Objets[i] = Temp;
            }
            m_Canevas.SetZone (null);
            MiseAJourPosition (me.getPoint ());
            m_Canevas.repaint ();
            break;
         default:
         };
      }
      public void mouseEntered (MouseEvent me) { } // TBD
      public void mouseExited (MouseEvent me) { } // TBD
   }


   //
   // Constructeurs
   //
   public Orga (final String Titre)
   {
      super (Titre);
      m_Objets = null;
      m_Canevas = null;
      m_Selection = null;
      m_Survol = null;
      m_BarreEtat = null;
      m_Forme = FORME_RECTANGLE;
      m_Mode = MODE_CREATION;
   }
   public Orga () { this (TITRE_DEFAUT); }
   //
   // Fermeture du programme (simili destructeur)
   //
   public void Fermer ()
   {
      System.exit (0);
   }


   private class BarreOutilsFormes
      extends JPanel implements ActionListener
   {
      private JButton BoutonRectangle, BoutonLosange, BoutonEllipse;
      public BarreOutilsFormes ()
      {
         BoutonRectangle = null;
         BoutonLosange = null;
         BoutonEllipse = null;
      }
      public void init ()
      {
         setLayout (new BorderLayout ());
         add (new JLabel ("Crer...?", JLabel.CENTER), BorderLayout.NORTH);

         JPanel PanneauTemporaire = new JPanel ();
         PanneauTemporaire.setLayout (new GridLayout (3, 1));
         PanneauTemporaire.add (BoutonRectangle = new JButton ("Rectangle"));
         PanneauTemporaire.add (BoutonLosange = new JButton ("Losange"));
         PanneauTemporaire.add (BoutonEllipse = new JButton ("Ellipse"));
         BoutonRectangle.addActionListener (this);
         BoutonLosange.addActionListener (this);
         BoutonEllipse.addActionListener (this);
         add (PanneauTemporaire, BorderLayout.CENTER);
         MettreAJour ();
      }

      public void MettreAJour ()
      {
         switch (m_Forme)
         {
         case FORME_RECTANGLE:
            BoutonRectangle.setEnabled (false);
            BoutonLosange.setEnabled (true);
            BoutonEllipse.setEnabled (true);
            break;
         case FORME_LOSANGE:
            BoutonRectangle.setEnabled (true);
            BoutonLosange.setEnabled (false);
            BoutonEllipse.setEnabled (true);
            break;
         case FORME_ELLIPSE:
            BoutonRectangle.setEnabled (true);
            BoutonLosange.setEnabled (true);
            BoutonEllipse.setEnabled (false);
            break;
         default:
         }
         m_BarreEtat.SetForme (m_Forme);
      }

      public void RendreIndisponible ()
      {
         BoutonRectangle.setEnabled (false);
         BoutonLosange.setEnabled (false);
         BoutonEllipse.setEnabled (false);
         m_BarreEtat.ResetForme ();
      }

      public void actionPerformed (ActionEvent ae)
      {
         if (ae.getSource () == BoutonRectangle)
            m_Forme = FORME_RECTANGLE;
         else if (ae.getSource () == BoutonLosange)
            m_Forme = FORME_LOSANGE;
         else if (ae.getSource () == BoutonEllipse)
            m_Forme = FORME_ELLIPSE;
         MettreAJour ();
      }
   }

   private class BarreOutilsActions
      extends JPanel implements ActionListener
   {
      private JButton
         BoutonSelectionner,
         BoutonCreer,
         BoutonDeplacer,
         BoutonDevant,
         BoutonArriere;
      public BarreOutilsActions ()
      {
         BoutonSelectionner = null;
         BoutonCreer = null;
         BoutonDeplacer = null;
         BoutonDevant = null;
         BoutonArriere = null;
         setLayout (new GridLayout (5, 1));
      }
      public void init ()
      {
         add (BoutonSelectionner = new JButton ("Slectionner"));
         add (BoutonCreer = new JButton ("Crer"));
         add (BoutonDeplacer = new JButton ("Dplacer"));
         add (BoutonDevant = new JButton ("Au devant"));
         add (BoutonArriere = new JButton (" l'arrire"));
         BoutonSelectionner.addActionListener (this);
         BoutonCreer.addActionListener (this);
         BoutonDeplacer.addActionListener (this);
         BoutonDevant.addActionListener (this);
         BoutonArriere.addActionListener (this);
         MettreAJour ();
      }

      public void MettreAJour ()
      {
         switch (m_Mode)
         {
         case MODE_SELECTION:
            BoutonSelectionner.setEnabled (false);
            BoutonCreer.setEnabled (true);
            BoutonDeplacer.setEnabled (true);
            BoutonDevant.setEnabled (true);
            BoutonArriere.setEnabled (true);
            m_BarreOutilsFormes.RendreIndisponible ();
            break;
         case MODE_CREATION:
            BoutonSelectionner.setEnabled (true);
            BoutonCreer.setEnabled (false);
            BoutonDeplacer.setEnabled (true);
            BoutonDevant.setEnabled (true);
            BoutonArriere.setEnabled (true);
            m_BarreOutilsFormes.MettreAJour ();
            break;
         case MODE_DEPLACEMENT:
            BoutonSelectionner.setEnabled (true);
            BoutonCreer.setEnabled (true);
            BoutonDeplacer.setEnabled (false);
            BoutonDevant.setEnabled (true);
            BoutonArriere.setEnabled (true);
            m_BarreOutilsFormes.RendreIndisponible ();
            break;
         case MODE_PLACERDEVANT:
            BoutonSelectionner.setEnabled (true);
            BoutonCreer.setEnabled (true);
            BoutonDeplacer.setEnabled (true);
            BoutonDevant.setEnabled (false);
            BoutonArriere.setEnabled (true);
            m_BarreOutilsFormes.RendreIndisponible ();
            break;
         case MODE_PLACERDERRIERE:
            BoutonSelectionner.setEnabled (true);
            BoutonCreer.setEnabled (true);
            BoutonDeplacer.setEnabled (true);
            BoutonDevant.setEnabled (true);
            BoutonArriere.setEnabled (false);
            m_BarreOutilsFormes.RendreIndisponible ();
            break;
         default:
         }
         m_BarreEtat.SetMode (m_Mode);
      }

      public void actionPerformed (ActionEvent ae)
      {
         if (ae.getSource () == BoutonSelectionner)
            m_Mode = MODE_SELECTION;
         else if (ae.getSource () == BoutonCreer)
            m_Mode = MODE_CREATION;
         else if (ae.getSource () == BoutonDeplacer)
            m_Mode = MODE_DEPLACEMENT;
         else if (ae.getSource () == BoutonDevant)
            m_Mode = MODE_PLACERDEVANT;
         else if (ae.getSource () == BoutonArriere)
            m_Mode = MODE_PLACERDERRIERE;
         MettreAJour ();
      }
   }

   private class BarreEtat extends JPanel
   {
      private JLabel m_lblMode,
                     m_lblForme;
      public BarreEtat ()
      {
         m_lblMode = null;
         m_lblForme = null;
         setLayout (new FlowLayout ());
      }
      public void init ()
      {
         add (m_lblMode = new JLabel ("Mode: ???"));
         add (m_lblForme = new JLabel ("Forme: ???"));
      }
      public void SetMode (final int Mode)
      {
         final String
            TXTMODE_SELECTION      = "Slection",
            TXTMODE_CREATION       = "Cration",
            TXTMODE_DEPLACEMENT    = "Dplacement",
            TXTMODE_PLACERDEVANT   = "Placer devant",
            TXTMODE_PLACERDERRIERE = "Placer derrire";
         switch (Mode)
         {
         case MODE_SELECTION:
            m_lblMode.setText (TXTMODE_SELECTION);
            break;
         case MODE_CREATION:
            m_lblMode.setText (TXTMODE_CREATION);
            break;
         case MODE_DEPLACEMENT:
            m_lblMode.setText (TXTMODE_DEPLACEMENT);
            break;
         case MODE_PLACERDEVANT:
            m_lblMode.setText (TXTMODE_PLACERDEVANT);
            break;
         case MODE_PLACERDERRIERE:
            m_lblMode.setText (TXTMODE_PLACERDERRIERE);
            break;
         default:
         }
      }

      public void ResetForme ()
      {
         m_lblForme.setText ("");
      }

      public void SetForme (final int Forme)
      {
         final String
            TXTFORME_RECTANGLE = "Rectangle",
            TXTFORME_LOSANGE   = "Losange",
            TXTFORME_ELLIPSE   = "Ellipse";
         switch (Forme)
         {
         case FORME_RECTANGLE:
            m_lblForme.setText (TXTFORME_RECTANGLE);
            break;
         case FORME_LOSANGE:
            m_lblForme.setText (TXTFORME_LOSANGE);
            break;
         case FORME_ELLIPSE:
            m_lblForme.setText (TXTFORME_ELLIPSE);
            break;
         default:
         }
      }
   }

   //
   //
   //
   public void init ()
   {
      //
      // Crer le canevas
      //
      m_Objets = new IAffichable [1];
      m_Objets[0] = null;
      m_Canevas = new Canevas ();
      getContentPane ().add (m_Canevas, BorderLayout.CENTER);

      //
      // Crer la barre d'tat (totu de suite, car les autres s'en servent)
      //
      m_BarreEtat = new BarreEtat ();
      getContentPane ().add (m_BarreEtat, BorderLayout.SOUTH);
      m_BarreEtat.init ();

      //
      // Crer la barre d'outils pour les formes
      //

      m_BarreOutilsFormes = new BarreOutilsFormes ();
      getContentPane ().add (m_BarreOutilsFormes, BorderLayout.WEST);
      m_BarreOutilsFormes.init ();

      //
      // Crer la barre d'outils pour les actions
      //
      m_BarreOutilsActions = new BarreOutilsActions ();
      getContentPane ().add (m_BarreOutilsActions, BorderLayout.EAST);
      m_BarreOutilsActions.init ();

      //
      // Pour assurer la bonne fermeture du programme
      //
      addWindowListener
      (
         new WindowAdapter () { public void windowClosing (WindowEvent we) { Fermer(); } }
      );


      //
      // Pour suivre la souris
      //
      LimierSouris ls = new LimierSouris ();
      GestionnaireSouris gs = new GestionnaireSouris ();
      m_Canevas.addMouseListener (gs);
      m_Canevas.addMouseMotionListener (gs);
      m_Canevas.addMouseListener (ls);
      m_Canevas.addMouseMotionListener (ls);
      ls.init ();
      ls.show ();
   }

   //
   //
   //
   private class Canevas extends JPanel
   {
      private Polygon m_Zone;

      public Canevas ()
      {
         m_Zone = null;
         setBackground (Color.black);
      }
      void SetZone (Polygon p)
      {
         m_Zone = p;
         if (m_Mode != MODE_SELECTION) return;
         m_Selection = new IAffichable [m_Objets.length];
         int NbDedans = 0;
         if (m_Zone != null)
         {
            int i;
            for (i = 0; i < m_Objets.length; i++)
            {
               if (m_Objets[i] != null)
               {
                  if (m_Zone.contains (m_Objets[i].GetBoundingBox ()))
                  {
                     m_Selection[NbDedans ++] = m_Objets[i];
                     m_Objets[i].Selectionner ();
                  }
                  else
                  {
                     m_Objets[i].Deselectionner ();
                  }
               }
            }
            for (i = NbDedans; i < m_Selection.length; i++)
            {
               m_Selection[i] = null;
            }
         }
         if (NbDedans == 0) m_Selection = null;
      }
      Polygon GetZone () { return m_Zone; }

      public void paint (Graphics g)
      {
         g.clearRect (0, 0, (int) getSize().getWidth (), (int) getSize().getHeight () );
         super.paint (g);
         for (int i = 0; i < m_Objets.length; i++)
         {
            if (m_Objets[i] != null)
            {
               m_Objets[i].Dessiner (g);
            }
         }
         g.setColor (Color.red);
         if (GetZone () != null) g.drawPolygon (GetZone ());
      }
   }

   //
   //
   //
   public static void main (String [] args)
   {
      Orga p = new Orga ();
      p.init ();
      p.setSize (new Dimension (LARGEUR_DEFAUT, HAUTEUR_DEFAUT));
      p.show ();
   }
}